namespace X3DTK {
namespace MESH {

template<class MData, class VData, class EData, class FData, bool RW>
TemplateNormalsUpdaterVerticesCoreVisitor<MData, VData, EData, FData, RW>::TemplateNormalsUpdaterVerticesCoreVisitor()
{
  define(Recorder<TemplateVertex<MData, VData, EData, FData, RW> >::getEnterFunction(&TemplateNormalsUpdaterVerticesCoreVisitor::enterVertex));
  define(Recorder<Shape>::getWalkOnFunction(&TemplateNormalsUpdaterVerticesCoreVisitor::walkOnShape));
}

template<class MData, class VData, class EData, class FData, bool RW>
bool TemplateNormalsUpdaterVerticesCoreVisitor<MData, VData, EData, FData, RW>::walkOnShape(Shape *, SFNode child)
{
  TemplateNormalsUpdaterStateVariables<MData, VData, EData, FData, RW> *stateVariables = Singleton<TemplateNormalsUpdaterStateVariables<MData, VData, EData, FData, RW> >::getInstance();
  MFNode selectedNodes = stateVariables->getSelectedNodes();
  return (selectedNodes.empty() || (std::find(selectedNodes.begin(), selectedNodes.end(), child) != selectedNodes.end()));
}

template<class MData, class VData, class EData, class FData, bool RW>
void TemplateNormalsUpdaterVerticesCoreVisitor<MData, VData, EData, FData, RW>::enterVertex(TemplateVertex<MData, VData, EData, FData, RW> *M)
{
  #ifdef TEMPLATE_SPECIALIZATION_SUPPORTED
  if (VData::template find<MESH::VertexNormalData>() 
   && FData::template find<MESH::FaceNormalData>())
   #endif
  { 
    const typename TemplateVertex<MData, VData, EData, FData, RW>::MFVertex &vertices = M->getVertices();
    for (typename TemplateVertex<MData, VData, EData, FData, RW>::MFVertex::const_iterator v = vertices.begin(); v != vertices.end(); ++v)
    {
      SFVec3f normal;

      // Compute point normal as an area weighted average of faces' normals
      const typename SFTemplateVertex<MData, VData, EData, FData, RW>::MFFace &faces = (*v)->getFaces();
      typename SFTemplateVertex<MData, VData, EData, FData, RW>::MFFace::const_iterator f = faces.begin();

      while (f != faces.end())
      {
        #ifdef TEMPLATE_SPECIALIZATION_SUPPORTED
        normal = normal + (*f)->template ogetData<MESH::FaceNormalData>().getNormal();
        #else
        normal = normal + (*f)->data().getNormal();
        #endif
        
        ++f;
      }  
      normal.normalize();

      // Store computed normal in VertexNormalData
      f = faces.begin();
      
      while (f != faces.end())
      {  
        #ifdef TEMPLATE_SPECIALIZATION_SUPPORTED
        (*v)->template ogetData<MESH::VertexNormalData>().setNormalOfFace(normal, *f);
        #else
        (*v)->data().setNormalOfFace(normal, *f);
        #endif
        
        ++f;
      } 
    }
  }
}

}
}
